import os
import time
import re

from unity_build_pipeline.Modes.modes_base.command import BaseCommand
from unity_build_pipeline.Modes.modes_base.base_modes import ParserEndpoint
from unity_build_pipeline.Services.Project import Project
from unity_build_pipeline.Services.Unity import Unity
from unity_build_pipeline.Support.config import pipeline_source_path
from unity_build_pipeline.Support.fileutils import copy, search_recursive
from unity_build_pipeline.Support.logger import color_print, RED, YELLOW, GREEN
from unity_build_pipeline.Support.shell import question_yes_no, quiz, question
from unity_build_pipeline.Support.shell import run

pattern = r"\"?([-a-zA-Z0-9.`?{}]+@\w+\.\w+)\"?"


def remove_prefix(string: str, prefix, count=1):
    if string.startswith(prefix):
        string = string.replace(prefix, '', count)
    return string


class InitMode(BaseCommand):
    def __init__(self, kwargs):
        super().__init__(kwargs)
        self.forced = kwargs.get('force', False)

    def run(self):
        if not self._check_all():
            return

        unity = Unity.detect_version(self.project_path)
        xcode_project_path = self._detect_xcode_project_path()
        if not unity or not xcode_project_path:
            return color_print("Project cannot be initialized", RED)

        use_backups = question_yes_no("Enable exported project backups?")

        ptrn = re.compile(pattern)
        username = question(
            "Please, provide your appleID (email) for further usage by fastlane",
            example="kek@gmail.com",
            check_answer=lambda p: re.match(ptrn, p),
            incorrect_answer=lambda p: "Email p %s is invalid" % p
        )

        bundleID = question(
            "Please, provide your app bundle id for further usage by fastlane",
            example="com.default.test"
        )

        teamID = question(
            "Please, provide your apple team id for further usage by fastlane",
            example="7NS7GL82HF"
        )

        project = Project.factory(
            self.project_path, unity, xcode_project_path, use_backups, username, bundleID, teamID)
        project.save()
        try:
            self._copy_stubs(project)
            color_print("Project successfully initialized!")
        except Exception as e:
            project.drop()
            color_print("Project initialization fail", RED)
            raise e

    def _check_all(self):
        if not self._check_project_folder():
            color_print("Project path %s is not exist or not writable" %
                        self.project_path, RED)
            return False
        if not self._check_if_unity_project():
            color_print("No Unity project detected in %s" %
                        self.project_path, RED)
            return False
        if Project.is_initialized(self.project_path):
            if not self.forced:
                color_print(
                    "Project already initialized. Use --force flag to reinitialize", RED)
                return False
            else:
                color_print(
                    "Project already initialized, but you've used --force flag!", YELLOW)
                project = Project(self.project_path)
                file_name = f":backup_before_init_{time.time()}"
                project.backup(Project.BUILD_TYPE_XCODE, file_name)
        return True

    def _check_project_folder(self):
        is_dir = os.path.isdir(self.project_path)
        accessable = os.access(self.project_path, os.W_OK)
        return is_dir and accessable

    def _check_if_unity_project(self):
        unity_path = self.project_path + Unity.PROJECT_VERSION_PATH
        return os.path.exists(unity_path)

    def _copy_stubs(self, project):
        old_config_path = pipeline_source_path
        new_config_path = project.get_config_folder(self.project_path)
        copy(old_config_path + '/stubs',
             new_config_path + '/stubs')

    def _detect_xcode_project_path(self):
        color_print("Searching for previous Xcode project...", GREEN)
        results = search_recursive(
            self.project_path, Project.XCODE_PROJECT_NAME)

        if not self.forced and len(results) > 0:
            if len(results) > 1:
                directory = quiz(
                    "Found multiple Xcode projects, which project to use?", results)
            else:
                directory = results.pop()
        else:
            directory = question(
                "Please, provide target ios project directory in project root",
                default_answer="Builds/IOS"
            ).strip()

        directory = remove_prefix(directory, self.project_path).lstrip('/')
        return directory


class InitEndpoint(ParserEndpoint):
    _parser_name = 'init'

    def run(self, kwargs):
        InitMode(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='Initialize project')
        parser.add_argument('-f', '--force', action='store_true')
        return parser



import os
from os.path import join
from string import Template

import yaml

from unity_build_pipeline.Support.fileutils import remove, copy, makedirs, search_recursive
from unity_build_pipeline.Support.logger import color_print, YELLOW, GREEN
from .Unity import Unity


class Project:
    BUILD_TYPE_XCODE = 'xcode'

    XCODE_PROJECT_NAME = 'Unity-iPhone.xcodeproj'

    def __init__(self, path, load=True):
        self.path = path
        self.unity = None
        self.backups_enabled = True
        self.build = {}
        self.commands = {}
        self.username = ""
        self.bundleID = ""
        self.teamID = ""
        if load:
            self.load()

    @staticmethod
    def is_initialized(path):
        return os.path.exists(Project.get_pipeline_path(path))

    @staticmethod
    def get_pipeline_path(path):
        return Project.get_config_folder(path) + '/config.yml'

    @staticmethod
    def get_config_folder(path):
        return path + '/.unity_build_pipeline'

    def get_stubs_folder(self):
        return self.get_config_folder(self.path) + '/stubs'

    @staticmethod
    def factory(project_path, unity, xcode_project_path, use_backups, username, bundle_id, team_id):
        project = Project(project_path, load=False)
        project.unity = unity
        project.build = {
            Project.BUILD_TYPE_XCODE: {
                'path': xcode_project_path,
            }
        }
        project.backups_enabled = use_backups
        project.username = username
        project.bundleID = bundle_id
        project.teamID = team_id
        project.commands = {}
        return project

    def load(self):
        config = yaml.safe_load(
            open(Project.get_pipeline_path(self.path), 'r').read())
        self.unity = Unity(config['unity']['version'], config['unity']['path'])
        self.build = {
            Project.BUILD_TYPE_XCODE: config['xcode_build']
        }
        self.backups_enabled = config['backups']
        self.commands = config['commands']
        self.username = config['username']
        self.bundleID = config['bundleID']
        self.teamID = config['teamID']

    def save(self):
        if Project.is_initialized(self.path):
            config = yaml.safe_load(
                open(Project.get_pipeline_path(self.path), 'r').read())
        else:
            makedirs(self.get_config_folder(self.path))
            config = {}
        config.update({
            'unity': {
                'version': self.unity.version,
                'path': self.unity.path,
            },
            'backups': self.backups_enabled,
            'xcode_build': {
                'path': self.get_export_path(Project.BUILD_TYPE_XCODE, absolute=False),
            },
            'commands': self.commands,
            'username': self.username,
            'bundleID': self.bundleID,
            'teamID': self.teamID
        })
        open(Project.get_pipeline_path(self.path), 'w').write(
            yaml.dump(config, default_flow_style=False))

    def drop(self):
        config_folder = Project.get_pipeline_path(self.path)
        remove(config_folder)

    def backup(self, build_type, suffix=':backup'):
        if not self.backups_enabled:
            return
        backup_path = self.get_export_path(build_type) + suffix
        if os.path.exists(self.get_export_path(build_type)):
            copy(self.get_export_path(build_type), backup_path)
            color_print("Project has been backed up in %s" %
                        backup_path, GREEN)
            return backup_path

    def restore_from_backup(self, build_type, remove_backup=True, suffix=':backup'):
        if not self.backups_enabled:
            return
        export_path = self.get_export_path(build_type)
        backup_path = export_path + suffix
        if os.path.exists(backup_path):
            remove(export_path)
            copy(backup_path, export_path)
            if remove_backup:
                remove(backup_path)
            color_print("Project has been restored from backup", GREEN)

    def get_export_path(self, build_type=BUILD_TYPE_XCODE, absolute=True):
        rel_part = self.build[build_type]['path']
        if absolute:
            return join(self.path, rel_part)
        return rel_part
        # return (self.path + '/' if absolute else '') + self.build[build_type]['path']

    def is_exported_project_valid(self, build_type):
        return os.path.exists(self.get_export_path(build_type) + '/' + Project.XCODE_PROJECT_NAME)

    def detect_exported_project(self, build_type):
        if build_type == Project.BUILD_TYPE_XCODE:
            return search_recursive(self.path, Project.XCODE_PROJECT_NAME)

    def export(self, allow_debugging, is_development, is_append_build=False, build_type=BUILD_TYPE_XCODE):
        success = False
        try:
            if not self.unity.is_build_type_supported(build_type):
                raise Exception("%s build is not supported by %s version of Unity" % (
                    build_type, self.unity.version))
            color_print('Exporting %s project..' % build_type)
            export_path = self.get_export_path(build_type)
            log_file = export_path + '/export.log'
            command = self._generate_export_command(
                allow_debugging, is_development, build_type, is_append_build)
            if command:
                color_print(
                    "Please, be patient. You can follow log here %s\n" % log_file, YELLOW)
                success = self.unity.execute(
                    self.path,
                    ["-batchmode", "-quit", "-executeMethod",
                        command, "-logFile", log_file]
                )
                if success:
                    color_print(
                        "Project has been successfully exported!", GREEN)
                else:
                    raise Exception(
                        "Unity export failed! Check logs here: %s" % log_file)
        finally:
            self._cleanup(build_type)
        return success

    def _generate_export_command(self, allow_debugging, is_development, build_type, is_append_build=False):
        if build_type == Project.BUILD_TYPE_XCODE:
            color_print('Writing export tool file..')
            makedirs(self.path + '/Assets/Editor/')
            with open(self.get_config_folder(self.path) + '/stubs/Unity/IOSExportTool.cs.stub', 'r') as stub_file:
                stub = Template(stub_file.read())
                stub = stub.substitute(
                    allowDebugging=('true' if allow_debugging else 'false'),
                    isDevelopment=('true' if is_development else 'false'),
                    buildOptions=(
                        'BuildOptions.None' if not is_append_build else 'BuildOptions.AcceptExternalModificationsToPlayer'),
                    buildPath=self.get_export_path(Project.BUILD_TYPE_XCODE)
                )
                with open(self.path + '/Assets/Editor/IOSExportTool.cs', 'w') as export_tool_file:
                    export_tool_file.write(stub)
                    return 'IOSExportTool.ExportXcodeProject'
        raise Exception("Exporting %s project is not supported" % build_type)

    def _cleanup(self, build_type):
        if build_type == Project.BUILD_TYPE_XCODE:
            path = self.path + '/Assets/Editor/IOSExportTool.cs'
            remove(path)



import os

import yaml

from unity_build_pipeline.Modes.modes_base.command import InitedCommand
from unity_build_pipeline.Modes.modes_base.base_modes import ParserEndpoint
from unity_build_pipeline.Services.Project import Project
from unity_build_pipeline.Support.logger import color_print, RED
from unity_build_pipeline.Support.shell import run


class RunMode(InitedCommand):
    def __init__(self, kwargs):
        super().__init__(kwargs)
        self.command = kwargs['command']
        self.options = kwargs['options']

    def run(self):
        project = Project(self.project_path)

        if self.command in ['list', 'ls']:
            print(yaml.dump(project.commands, default_flow_style=False))
            return

        if self.command not in project.commands:
            color_print("Command %s is not defined for this project" % self.command, RED)

        cmd = project.commands[self.command]
        if isinstance(cmd, str):
            shell_cmd = self._format(project, cmd)
            cwd = os.getcwd()
            run(shell_cmd.split(" "), cwd)
        else:
            commands = [cmd['command']] if isinstance(cmd['command'], str) else cmd['command']
            for command in commands:
                shell_cmd = self._format(project, command)
                cwd = self._format(project, cmd['cwd']) if 'cwd' in cmd else os.getcwd()
                run(shell_cmd.split(" "), os.path.abspath(os.path.expanduser(cwd)))

    def _format(self, project, string):
        vars = {
            'cwd': os.getcwd(),
            'unity_version': project.unity.version,
            'unity_path': project.unity.path,
            'project_path': project.path,
            'xcode_build_path': project.get_export_path(Project.BUILD_TYPE_XCODE, absolute=False),
            'xcode_build_path_absolute': project.get_export_path(Project.BUILD_TYPE_XCODE, absolute=True)
        }
        try:
            return string.format(*tuple(self.options), **vars)
        except IndexError as e:
            color_print("Cannot format \"%s\" string from command. Missing some options" % string, RED)
            exit(0)

class RunModeEndpoint(ParserEndpoint):
    _parser_name = 'run'

    def run(self, kwargs):
        RunMode(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='Run command')
        parser.add_argument('command', help='command to execute')
        parser.add_argument('options', nargs='*')
        return parser



from unity_build_pipeline.Modes.modes_base.base_modes import ParserEndpoint, ParserNode
from unity_build_pipeline.Modes.modes_base.command import InitedCommand, BaseCommand
from unity_build_pipeline.Modes.pipeline_modes.InitMode import InitMode


class CommandA(InitedCommand):  # inited: like base but checks if project is inited
    def run(self):
        print(*self.kwargs)


class CommandB(BaseCommand):
    def run(self):
        print("test", self.kwargs.get("testArgs", {}))


class CommandAB(BaseCommand):
    def run(self):
        kwargs = self.kwargs
        CommandA(kwargs).start()
        CommandB(kwargs).start()


class AEndpoint(ParserEndpoint):  # endpoint is a "leaf" of command parsing
    _parser_name = 'A'

    def run(self, kwargs):
        CommandA(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='Run A command')
        return parser


class BEndpoint(ParserEndpoint):
    _parser_name = 'B'

    def run(self, kwargs):
        CommandB(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='Run B command')
        parser.add_argument('honkyArgs', nargs='*',
                            help='args to be flowed to B command')
        return parser


class ABEndpoint(ParserEndpoint):
    _parser_name = 'AB'

    def run(self, kwargs):
        InitMode(kwargs).start()
        CommandAB(kwargs).start()
        # CommandA(kwargs).start()
        # CommandB(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='Run both A and B commands')
        parser.add_argument('honkyArgs', nargs='*',
                            help='args to be flowed to B command')
        return parser


class NestedExampleNode(ParserNode):  # node is a
    _parser_name = 'neEx2'
    # required for a node, shouldn't be repeated in children. Can be any string
    _children_name = 'nested_example_modes2'
    # list of nodes/endpoints accessed from this node
    _children_classes = [AEndpoint, BEndpoint, ABEndpoint]
    extra_parser_args = dict(help="Nested example command node")
    # ˆ additional kwargs for parent's subparsers.add_parser()
    # See https://docs.python.org/3/library/argparse.html#sub-commands

    def run(self, kwargs):
        print("/-- Nested step run --/")  # additional stuff
        super().run(kwargs)  # node's super() delegates job to children if they're called


class ExampleNode(ParserNode):
    _parser_name = 'neEx'
    _children_name = 'nested_example_modes'
    _children_classes = [NestedExampleNode, BEndpoint]
    extra_subparser_args = dict(required=True)
    extra_parser_args = dict(help="Root-node for examples")

# exampleNode is added to Root class'' _children_classes
# Root is ParserNode too but a bit trickier (to be actually "root")



import os
import subprocess

from unity_build_pipeline.Support.logger import color_raw_input, color_print, RESET, YELLOW, RED, GREEN


def run(cmd, cwd=None, silent=False):
    if cwd is None:
        cwd = os.getcwd()
    if silent:
        return try_run(cmd, cwd, silent)
    else:
        color_print("Running command:", YELLOW)
        color_print(' '.join(cmd), YELLOW)
        exit_code = try_run(cmd, cwd, silent)
        if exit_code == 0:
            color_print("Command success!", GREEN)
        else:
            color_print(f"Command failed with code {exit_code}!", RED)
        return exit_code


def try_run(cmd, cwd, silent=False):
    try:
        stdout = subprocess.DEVNULL if silent else None
        exit_code = subprocess.check_call(cmd, cwd=cwd, stdout=stdout)
        return exit_code
    except subprocess.CalledProcessError as e:
        exit_code = e.returncode
        return exit_code


def question_yes_no(q, default_answer='y', color=RESET):
    answer = question(q, default_answer=('y' if default_answer else 'n'), answers=['y', 'n'], color=color)
    return answer == 'y'


def question(question, check_answer=None, incorrect_answer=None, example=None, default_answer=None, answers=None,
             color=RESET):
    def def_check_answer(ans):
        if ans == '' and default_answer is not None:
            return True
        if ans.upper() in answers or ans in answers or ans.lower in answers:
            return True
        return False

    if answers:
        if check_answer is None:
            check_answer = def_check_answer
        answers = list(map(lambda x: x.upper() if x == default_answer else x, answers))
        answers_help = ' [%s]' % '/'.join(answers)
    else:
        if default_answer:
            answers_help = ' [%s]' % default_answer
        else:
            answers_help = ''
    if example:
        answers_help = " (ex. %s)%s" % (example, answers_help)
    actual_question = question + answers_help + ": "

    actual_answer = color_raw_input(actual_question, color)
    if check_answer:
        if incorrect_answer is None:
            incorrect_answer = lambda ans: 'Type one of available answers: %s' % ", ".join(answers)
        answer_correct = check_answer(actual_answer)
        while not answer_correct:
            if incorrect_answer:
                color_print(incorrect_answer(actual_answer), YELLOW)
            actual_answer = color_raw_input(actual_question, color)
            answer_correct = check_answer(actual_answer)
    else:
        answer_correct = True
    if actual_answer == '' and default_answer and answer_correct:
        return default_answer.strip()
    return actual_answer.strip()

class Question:
    @staticmethod
    def _const_false():
        return False

    @classmethod
    def checked_once(cls, question_func, check_func):
        return cls.checked(
            question_func=question_func,
            check_func=check_func,
            retry_func=cls._const_false(),
            retries=1
        )


    @staticmethod
    def decorated_ask_continue(message='', color=RESET):
        message = message or "Command failed; retry?"
        return lambda: question_yes_no(message, color=color)

    @classmethod
    def retry_while_wanna(cls, question_func, check_func, message='', color=RED):
        return cls.checked(
            question_func=question_func,
            check_func=check_func,
            retry_func=cls.decorated_ask_continue(message, color),
        )

    @classmethod
    def checked(cls, question_func, check_func, retry_func, retries=100):
        try_num = 0
        while True:
            try_num += 1
            ans = question_func()
            if check_func(ans):
                return True
            elif try_num >= retries or not retry_func(ans):
                return False



def quiz(q, answers, color=RESET):
    color_print(q, color)
    lst = []

    def __print_variants(ans):
        for i, p in enumerate(answers):
            color_print("[%s] %s" % (i + 1, p,), color)
            lst.append(str(i + 1))
        return '' if ans is None else "\"%s\" is not one of the options numbers!" % ans

    __print_variants(None)
    idx = int(question(
        'Choose one of the options above (Enter the number of the selected option)',
        answers=lst, incorrect_answer=__print_variants, color=color))
    return answers[idx - 1]



GREEN = '\x1b[32m'
YELLOW = '\033[93m'
RED = '\033[91m'
WHITE = '\x1b[32m'
RESET = '\x1b[0m'

def colority_string(string, color=RESET):
    return color + string + RESET


def color_print(string, color=RESET):
    print(colority_string(string, color))


def color_raw_input(prompt, color=RESET):
    return input(colority_string(prompt, color)).strip()



import time

from unity_build_pipeline.Modes.modes_base.base_modes import ParserEndpoint, ParserNode
from unity_build_pipeline.Modes.modes_base.command import InitedCommand, BaseCommand
from unity_build_pipeline.Modes.pipeline_modes.InitMode import InitMode
from unity_build_pipeline.Modes.pipeline_modes.ExportMode import ExportMode
from unity_build_pipeline.Modes.pipeline_modes.FastlaneMode import FastlaneMode
from unity_build_pipeline.Support.logger import color_print, WHITE

class RapidCommand(BaseCommand):
    def run(self):
        kwargs = self.kwargs
        start_time = time.time()
        
        InitMode(kwargs).start()
        ExportMode(kwargs).start()
        FastlaneMode(kwargs).start()

        elapsed_min = str(round((time.time() - start_time)/60))
        color_print("Pipeline just saved you " + elapsed_min + " min", WHITE)

class RapidEndpoint(ParserEndpoint):
    _parser_name = 'runall'

    def run(self, kwargs):
        RapidCommand(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='One command to run them all')
        parser.add_argument('-f', '--force', action='store_true')
        parser.add_argument('options', nargs='*', help='fastlane options')
        parser.add_argument('--allow-debugging', action='store_true',
                            help='Export with AllowDebugging option')
        parser.add_argument('--dev', action='store_true',
                            help='Export development project')
        return parser




import os
import time
from os.path import exists, join

from unity_build_pipeline.Modes.modes_base.command import InitedCommand
from unity_build_pipeline.Modes.modes_base.base_modes import ParserEndpoint
from unity_build_pipeline.Services.Project import Project
from unity_build_pipeline.Services.Unity import Unity
from unity_build_pipeline.Support.fileutils import remove, copy, makedirs
from unity_build_pipeline.Support.logger import color_print, GREEN, YELLOW
import plistlib as plist

migrate_fields = [
    'CFBundleIdentifier',
    'CFBundleName',
    'CFBundleDisplayName'
]


def get_info_plist_path(root):
    return root + '/Info.plist'


class ExportMode(InitedCommand):
    def __init__(self, kwargs):
        super().__init__(kwargs)
        self.export_params = {
            'allow_debugging': kwargs['allow_debugging'],
            'is_development': kwargs['dev'],
        }

    @staticmethod
    def configure_arg_parser(subparsers):
        parser = subparsers.add_parser('export', help='Export project')
        parser.add_argument('--allow-debugging', action='store_true', help='Export with AllowDebugging option')
        parser.add_argument('--dev', action='store_true', help='Export development project')
        return parser

    def run(self):
        project = Project(self.project_path)
        color_print("Exporting project", GREEN)
        detected_unity = Unity.detect_version(self.project_path)
        use_append_strategy = detected_unity.version == project.unity.version

        is_first_export = not os.path.exists(project.get_export_path(Project.BUILD_TYPE_XCODE))
        if is_first_export:
            color_print("No exported project detected. Performing clean export", YELLOW)
            return self._export(project, False)

        if not use_append_strategy:
            warning = "Detected Unity version difference was: {} now: {}" \
                .format(project.unity.version, detected_unity.version)
            color_print(warning, YELLOW)
            color_print("Using REPLACE strategy for project build")
        else:
            color_print("Using APPEND strategy for project build")
        self._export(project, use_append_strategy)

    def _get_meta_path(self, export_path):
        label = time.time()
        meta_files_path = join(export_path, f'_meta_{label}')
        return meta_files_path

    def _export(self, project, use_append_strategy):
        try:
            project.backup(build_type=Project.BUILD_TYPE_XCODE)
            export_path = project.get_export_path(Project.BUILD_TYPE_XCODE)
            meta_files_path = self._get_meta_path(export_path)

            self._backup_meta_files(export_path, meta_files_path)
            project.export(
                *self.export_params,
                build_type=Project.BUILD_TYPE_XCODE,
                is_append_build=use_append_strategy
            )
            if not use_append_strategy:
                self._migrate_plist(meta_files_path, export_path)
            remove(meta_files_path)
        except Exception as e:
            project.restore_from_backup(Project.BUILD_TYPE_XCODE)
            raise e

    def _backup_meta_files(self, src_root, dst_root):
        makedirs(dst_root)
        copy(src_root + '/Info.plist', dst_root)

    def _migrate_plist(self, src_root, dst_root):
        src_path = get_info_plist_path(src_root)
        dst_path = get_info_plist_path(dst_root)
        if not exists(src_root):
            return

        color_print("Settings migration..")

        src_plist = plist.load(open(src_path))
        dst_plist = plist.load(open(dst_path))

        self._replace_fields(src_plist, dst_plist, self._get_migrate_fields())
        plist.dump(dst_plist, open(dst_path, 'w'))

    def _replace_fields(self, plist_src, plist_dest, fields):
        for field_name in fields:
            plist_src[field_name] = plist_dest[field_name]

    def _get_migrate_fields(self):
        return migrate_fields

class ExportEndpoint(ParserEndpoint):
    _parser_name = 'export'
    extra_parser_args = dict()

    def run(self, kwargs):
        ExportMode(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='Export project')
        parser.add_argument('--allow-debugging', action='store_true', help='Export with AllowDebugging option')
        parser.add_argument('--dev', action='store_true', help='Export development project')
        return parser


import glob
import os
import re
from unity_build_pipeline.Support.logger import color_print, YELLOW
from unity_build_pipeline.Support.fileutils import search_recursive
from unity_build_pipeline.Support.shell import run, question


class Unity:
    MAC_UNITY_APP = 'Unity.app'
    MAC_EXECUTABLE_RELATIVE_PATH = '/Contents/MacOS/Unity'
    MAC_EXECUTABLE_NAME = '/' + MAC_UNITY_APP
    MAC_PLAYBACK_ENGINES_PATH = '/PlaybackEngines'
    PROJECT_VERSION_PATH = '/ProjectSettings/ProjectVersion.txt'
    MAC_UNITY_SEARCH_DIRS = [
        '/Applications/Unity/Hub/Editor/', '/Applications/Unity'
    ]

    @staticmethod
    def get_available_executables():
        executables = []
        for x in Unity.MAC_UNITY_SEARCH_DIRS:
            results = search_recursive(x, Unity.MAC_UNITY_APP)
            for res in results:
                executables.append(Unity(res.split('/').pop(), res))
        return executables

    @staticmethod
    def detect_version(project_path):
        with open(project_path + Unity.PROJECT_VERSION_PATH, 'r') as f:
            version = re.match(re.compile(".*(\d\d\d\d\.\d+\.\d+[a-z]+\d+).*"), f.read()).group(1)
            color_print("Detected Unity version %s" % version)
            unity_exec = None
            for executable in Unity.get_available_executables():
                if version == executable.version:
                    color_print("Unity executable found in %s" % executable.path)
                    return executable
            if unity_exec is None:
                color_print("Couldn't find Unity executable for version %s" % version, YELLOW)
                path = question(
                    "Please, provide path to Unity %s executable" % version,
                    example="/Applications/Unity/2018.2.20f1",
                    check_answer=lambda p: os.path.exists(p + Unity.MAC_EXECUTABLE_RELATIVE_PATH),
                    incorrect_answer=lambda p: "Unity path %s is invalid" % p
                )
                if path.endswith("/Unity.app"):
                    path = path.replace("/Unity.app", "")
                unity_exec = Unity(version, path)
            return unity_exec

    def __init__(self, version, path):
        self.version = version
        self.path = path.rstrip('/')

    def is_build_type_supported(self, build_type):
        map = {
            'xcode': 'iossupport'
        }
        for path in glob.glob(self.path + Unity.MAC_PLAYBACK_ENGINES_PATH + '/*Support'):
            if map[build_type] in path.lower():
                return True
        return False

    def execute(self, pwd, args):
        exit_code = run(
            [self.path + Unity.MAC_EXECUTABLE_NAME + Unity.MAC_EXECUTABLE_RELATIVE_PATH, '-projectPath', pwd] + args,
            cwd=pwd
        )
        return exit_code == 0



import argparse

from unity_build_pipeline.Modes.pipeline_modes.RapidMode import RapidEndpoint
from unity_build_pipeline.Modes.pipeline_modes.ExportMode import ExportEndpoint
from unity_build_pipeline.Modes.pipeline_modes.FastlaneMode import FastlaneEndpoint
from unity_build_pipeline.Modes.pipeline_modes.InitMode import InitEndpoint
from unity_build_pipeline.Modes.modes_base.base_modes import ParserNode
from unity_build_pipeline.Modes.pipeline_modes.RunMode import RunModeEndpoint


class Root(ParserNode):
    _parser_name = 'Pipeline'
    _children_name = 'pipeline_mode'
    extra_subparser_args = dict(required=True)
    _children_classes = [
        InitEndpoint,
        ExportEndpoint,
        FastlaneEndpoint,
        RunModeEndpoint,
        RapidEndpoint
    ]

    def __init__(self):
        super().__init__()
        self.parser = self.make_parser()

    def parse(self):
        return self.parser.parse_args().__dict__

    def make_parser(self):
        parser = argparse.ArgumentParser()
        sub = self.make_sub(parser)
        self.configure_children(sub)
        return parser



import fnmatch
import os
import shutil
from distutils.dir_util import copy_tree


def makedirs(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if os.path.isdir(path):
            pass
        else:
            raise

def replace_string_entries(filename, old_string, new_string):
    # Safely read the input filename using 'with'
    with open(filename) as f:
        s = f.read()
        if old_string not in s:
            # print('"{old_string}" not found in {filename}.'.format(**locals()))
            return

    # Safely write the changed content, if found in the file
    with open(filename, 'w') as f:
        # print('Changing "{old_string}" to "{new_string}" in {filename}'.format(**locals()))
        s = s.replace(old_string, new_string)
        f.write(s)


def copy(src, dst):
    if os.path.exists(src):
        makedirs(dst)
        if os.path.isdir(src):
            copy_tree(src, dst)
        else:
            shutil.copy(src, dst)


def remove(src):
    if os.path.exists(src):
        if os.path.isdir(src):
            shutil.rmtree(src)
        else:
            os.remove(src)


def rotate_dirs(dirs):
    delimiter = '::rotation::'
    dirs.sort(reverse=True)
    for dir in dirs:
        suffix_list = dir.split(delimiter)[1:]
        suffix = 0 if not suffix_list else int(suffix_list[0])
        dst = dir.replace(delimiter + str(suffix), '') + \
              delimiter + str(suffix + 1)
        shutil.move(dir, dst)


def search_recursive(root, pattern):
    result = []
    for root, dirnames, filenames in os.walk(root):
        if fnmatch.filter(dirnames, pattern):
            result.append(root.replace(root.rstrip('/') + '/', ''))
    return result



import os
from unity_build_pipeline.Services.Project import Project
from unity_build_pipeline.Support.logger import color_print, YELLOW, GREEN, RED
from unity_build_pipeline.Support.shell import question_yes_no


def init_project(project_path, kwargs):
    from ..pipeline_modes.InitMode import InitMode
    kwargs.update({'project_path': project_path})
    InitMode(kwargs).start()

class BaseCommand:
    op_name = None

    def __init__(self, kwargs):
        self.kwargs = kwargs
        self.project_path = os.getcwd().rstrip('/')

    def start(self):
        self._log_running(self.get_operation_name())
        self.run()

    def get_operation_name(self):
        return self.op_name or self.__class__.__name__

    def _log_running(self, operation_name):
        color_print(f'### Running {operation_name} ###', GREEN)

    def run(self):
        pass


class InitedCommand(BaseCommand):
    def start(self):
        if not self.check_inited():
            color_print("Canceled: isn't initialized.", color=RED)
            return
        super().start()

    def check_inited(self):
        if not Project.is_initialized(self.project_path):
            if question_yes_no("Project is not initialized. Initialize it now?", color=YELLOW):
                init_project(self.project_path, self.kwargs)
            else:
                return False
        return True



from unity_build_pipeline.Support.logger import color_print, GREEN
from unity_build_pipeline.Support.shell import run
from unity_build_pipeline.Support.fileutils import replace_string_entries


class Fastlane:
    def __init__(self, project):
        self.project = project

    def execute(self, args):
        project_path = self.project.get_export_path('xcode')
        self.ensure_install(project_path)
        color_print("Starting fastlane..", GREEN)
        run(['bundle', 'exec', 'fastlane'] + args, cwd=project_path)

    def ensure_install(self, path):
        color_print("Updating fastlane..", GREEN)
        run(['rm', '-rf', path + '/fastlane'], path)
        run(['rm', '-f', path + '/Gemfile'], path)
        run(['cp', '-R', self.project.get_stubs_folder() +
             '/Fastlane/fastlane', path + '/fastlane'], path)

        replace_string_entries(path + '/fastlane/Fastfile',
                               "options[:username]", "'"+self.project.username+"'")

        replace_string_entries(path + '/fastlane/Fastfile',
                               "options[:teamid]", "'" + self.project.teamID + "'")

        replace_string_entries(path + '/fastlane/Fastfile',
                               "options[:appid]", "'" + self.project.bundleID + "'")

        replace_string_entries(path + '/fastlane/Appfile',
                               "username", self.project.username)

        replace_string_entries(path + '/fastlane/Appfile',
                               "appid", self.project.bundleID)

        replace_string_entries(path + '/fastlane/Appfile',
                               "teamid", self.project.teamID)

        run(['cp', '-R', self.project.get_stubs_folder() +
             '/Fastlane/Gemfile', path + '/Gemfile'], path)
        run(['bundle', 'update'], cwd=path, silent=True)
        pbx_project_path = path + '/Unity-iPhone.xcodeproj/project.pbxproj'

        content = open(pbx_project_path, 'r').read()
        if 'VERSIONING_SYSTEM = "apple-generic"' not in content:
            color_print("Patching versioning system..", GREEN)
            content = self.patch_pbx(content)
            open(pbx_project_path, 'w').write(content)

    def patch_pbx(self, content):
        pbx = content.split("\n")
        new_pbx = []
        for i, line in enumerate(pbx):
            new_pbx.append(line)
            if "COPY_PHASE_STRIP" in line:
                new_pbx.append('				CURRENT_PROJECT_VERSION = 0.1;')
            if "UNITY_SCRIPTING_BACKEND" in line:
                new_pbx.append('				VERSIONING_SYSTEM = "apple-generic";')
        return "\n".join(new_pbx)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="UnityBuildPipeline",
    version="0.0.5",
    author="MadCoder",
    author_email="madcoder39@gmail.com",
    description="Unity build pipeline for iOS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MadCoder39/UnityBuildPipelineiOS",
    packages=setuptools.find_packages(),
    install_requires=required,
    scripts=['pipeline'],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
    ],
)



from os.path import join


class Encoder:
    def __init__(self, encode_func):
        self._encode_func = encode_func

    def encode(self, conf):
        return [
            self._encode_func(key, val)
            for key, val in conf.items() if val
        ]


pair_encoder = Encoder(
    lambda k, v: f'{k} "{v}"'
)


class Appfile:
    app_id = ''
    apple_id = ''

    def to_text(self):
        lines = self._serialize_assigned()
        return '\n'.join(lines)

    def _serialize_assigned(self):
        lines = pair_encoder.encode(self._get_pairs())
        return lines

    def _get_pairs(self):
        return {
            'app_identifier': self.app_id,
            'apple_id': self.apple_id
        }

    def save(self, path, filename):
        with open(join(path, filename), 'w') as f:
            f.write(self.to_text())




class BaseParser:
    _parser_name = ''

    @property
    def parser_name(self) -> str:
        if self._parser_name:
            return self._parser_name
        raise NotImplementedError(f"{self.__class__.__name__}: parser_name not set")

    def run(self, kwargs):
        raise NotImplementedError

class ParserNode(BaseParser):
    _children_name = ''
    _children_classes = []
    extra_parser_args = {}
    extra_subparser_args = {}

    def __init__(self):
        self._children_parsers = {cls._parser_name: cls() for cls in self._children_classes}


    def run(self, kwargs):
        mode_name = kwargs.get(self.children_name, None)
        if mode_name:
            target_parser = self._children_parsers.get(mode_name, None)
            if target_parser:
                target_parser.run(kwargs)

    @property
    def children_name(self):
        if self._children_name:
            return self._children_name
        raise NotImplementedError("submodes key name not set")


    def configure_arg_parser(self, parent_subparsers):
        parser = self.add_as_subparser(parent_subparsers)
        sub = self.make_sub(parser)
        self.configure_children(sub)
        return parser

    def add_as_subparser(self, parent_sub):
        return parent_sub.add_parser(**dict(
            name=self.parser_name,
            **self.get_extra_parser_args()
        ))

    def make_sub(self, parser):
        return parser.add_subparsers(**dict(
            dest=self._children_name,
            **self.get_extra_subparser_args()
        ))

    def configure_children(self, subparsers):
        for child in self._children_parsers.values():
            child.configure_arg_parser(subparsers)

    def get_extra_subparser_args(self):
        return self.extra_subparser_args

    def get_extra_parser_args(self):
        return self.extra_parser_args

class ParserEndpoint(BaseParser):
    def configure_arg_parser(self, parent_subparsers):
        return parent_subparsers.add_parser(self.parser_name)

    def run(self, kwargs):
        raise NotImplementedError


from unity_build_pipeline.Modes.modes_base.command import InitedCommand
from unity_build_pipeline.Modes.modes_base.base_modes import ParserEndpoint
from unity_build_pipeline.Services.Fastlane import Fastlane
from unity_build_pipeline.Services.Project import Project


class FastlaneMode(InitedCommand):
    def __init__(self, kwargs):
        super().__init__(kwargs)
        self.options = kwargs.get('options', list())

    def run(self):
        project = Project(self.project_path)
        fastlane = Fastlane(project)
        fastlane.execute(self.options)


class FastlaneEndpoint(ParserEndpoint):
    _parser_name = 'fastlane'

    def run(self, kwargs):
        FastlaneMode(kwargs).start()

    def configure_arg_parser(self, parent_subparsers):
        parser = parent_subparsers.add_parser(
            name=self.parser_name, help='Make project')
        parser.add_argument('options', nargs='*', help='fastlane options')
        return parser
